import math

from texasholdem import TexasHoldEm
from texasholdem.card.card import Card
from texasholdem.card.deck import Deck
from texasholdem.game.player_state import PlayerState
from texasholdem.game.action_type import ActionType
from texasholdem.game.hand_phase import HandPhase
from texasholdem.evaluator.evaluator import evaluate
from texasholdem.gui import TextGUI
import time

game = TexasHoldEm(500, 5, 2, max_players=6)
gui = TextGUI()


def _calculate_equity(hand: list[Card], board: list[Card], num_villians: int):
    board_copy = list(board)
    villian_hands: list[list[Card]] = []

    deck = Deck()
    while len(villian_hands) < num_villians:
        villian_hand = []
        while len(villian_hand) < 2:
            c = deck.draw()[0]
            if c not in hand and \
               c not in board_copy and \
               all(c not in v_hand for v_hand in villian_hands):
                villian_hand.append(c)
        villian_hands.append(villian_hand)

    deck = Deck()
    while len(board_copy) < 5:
        c = deck.draw()[0]
        if c not in hand and \
           c not in board_copy and \
           all(c not in v_hand for v_hand in villian_hands):
            board_copy.append(c)

    return 1 if all(evaluate(hand, board_copy) < evaluate(villian_hand, board_copy)
                    for villian_hand in villian_hands) else 0


def calculate_equity(hand: list[Card], board: list[Card], villian_hands: int, n: int = 25000):
    return sum(_calculate_equity(hand, board, villian_hands) for _ in range(n)) / n


def call_player(game: TexasHoldEm):
    player = game.players[game.current_player]
    if player.state == PlayerState.TO_CALL:
        return ActionType.CALL, None
    else:
        return ActionType.CHECK, None


def value_player(game: TexasHoldEm):
    player = game.players[game.current_player]

    chips_at_stake = sum(game._get_pot(i).get_total_amount() for i in range(player.last_pot + 1))
    pot_odds = game.chips_to_call(player.player_id) / chips_at_stake
    equity = calculate_equity(game.get_hand(player.player_id), game.board, len(list(game.active_iter())) - 1)
    value_bet = int(chips_at_stake * equity)

    print(f"pot_odds: {pot_odds}")
    print(f"equity: {equity}")
    print(f"value_bet: {value_bet}")

    if player.state == PlayerState.TO_CALL:
        raise_adjuster = 0.5 * equity - 0.5
        if game.hand_phase == HandPhase.PREFLOP:
            fold_adjuster = math.pow(equity, 1.3) + 0.2
        else:
            fold_adjuster = math.pow(equity, 1.3) - 0.2

        print(f"equity + fold_adjuster: {equity + fold_adjuster}")
        print(f"equity + raise_adjuster: {equity + raise_adjuster}")
        if equity + fold_adjuster < pot_odds:
            return ActionType.FOLD, None
        elif equity + raise_adjuster > pot_odds and value_bet >= 0.25 * chips_at_stake:
            value_bet = max(game.big_blind, value_bet)

            if player.chips <= value_bet:
                return ActionType.ALL_IN, None
            else:
                raise_amt = game.player_bet_amount(player.player_id) + game.chips_to_call(player.player_id) + value_bet
                return ActionType.RAISE, min(raise_amt, player.chips + game.player_bet_amount(player.player_id))
        else:
            return ActionType.CALL, None
    else:
        if equity > pot_odds + 0.05 and value_bet >= 0.25 * chips_at_stake:
            if player.chips <= value_bet:
                return ActionType.ALL_IN, None
            elif value_bet < game.big_blind:
                return ActionType.CHECK, None
            else:
                raise_amt = game.player_bet_amount(player.player_id) + game.chips_to_call(player.player_id) + value_bet
                return ActionType.RAISE, min(raise_amt, player.chips + game.player_bet_amount(player.player_id))
        else:
            return ActionType.CHECK, None


user_player_id = 5
gui.set_player_ids(range(game.max_players))

while game.is_game_running():
    game.start_hand()
    while game.is_hand_running():
        gui.print_state(game)

        if game.hand_phase != HandPhase.SETTLE:
            action_type, val = value_player(game)
            gui.print_action(game.current_player, action_type, val)
            game.take_action(action_type, val)
        else:
            game.take_action(ActionType.CHECK, None)



"""
Main package for texasholdem
"""

from texasholdem.game.game import TexasHoldEm
from texasholdem.game.hand_phase import HandPhase
from texasholdem.game.action_type import ActionType
from texasholdem.game.player_state import PlayerState



"""
The game module includes lightweight data classes:
- Player
    - The player class keeps track of the chips, player_state,
      last_pot, and player_id of the player.
- Pot
    - The pot class represents a betting pot which players
      can post to. Includes helper functions and attributes that deal
      with split pots or how much a player needs to call.

It also includes the main TexasHoldEm class of the texasholdem package.
"""

from __future__ import annotations

from typing import Iterator, Callable, Dict, Tuple, Union, Optional
from enum import Enum, auto
import random

from texasholdem.card.card import Card
from texasholdem.card.deck import Deck
from texasholdem.game.history import PrehandHistory, BettingRoundHistory, PlayerAction
from texasholdem.game.action_type import ActionType
from texasholdem.game.hand_phase import HandPhase
from texasholdem.game.player_state import PlayerState
from texasholdem.evaluator import evaluator


class Player:
    # pylint: disable=too-few-public-methods
    """
    The TexasHoldEm object uses this Player as a bookkeeping mechanism for:
        - Chip count
        - PlayerState
        - Which pots the player is in
    """

    def __init__(self, player_id, chips):
        self.player_id = player_id
        self.chips = chips
        self.state = PlayerState.IN

        # invariant: last_pot is the newest pot that player is eligible for
        self.last_pot = 0


class Pot:
    """
    Class representing a betting pot, players will post to the pot.
    If players post more than the raised amount, we set the new raised amount.

    At the end of a betting round, the bets are consolidated and reset.
    """

    def __init__(self):
        self.amount = 0
        self.raised = 0
        self.player_amounts = {}

    def chips_to_call(self, player_id: int) -> int:
        """Returns the amount of chips to call for the given player.

        Args:
            player_id (int): The player_id of the player in the pot
        Returns:
            int: The amount the player needs to call to be in this pot
              (this is just self.raised if the player hasn't bet yet)

        """
        return self.raised - self.player_amounts.get(player_id, 0)

    def player_post(self, player_id: int, amount: int):
        """The given player posts amount into this pot. If player[player_id].amount > raised,
        sets new raised value.

        Arguments:
            player_id (int): The player_id of the player posting
            amount (int): The amount to post into this pot

        """
        self.player_amounts[player_id] = self.player_amounts.get(player_id, 0) + amount

        if self.player_amounts[player_id] > self.raised:
            self.raised = self.player_amounts[player_id]

    def get_player_amount(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int): Player player_id
        Returns:
            int: the amount the player has bet currently for this pot.

        """
        return self.player_amounts.get(player_id, 0)

    def players_in_pot(self) -> Iterator[int]:
        """
        Returns:
            Iterator[int]: An iterator over the player player_id's that have a stake in this pot.

        """

        return iter(self.player_amounts.keys())

    def collect_bets(self):
        """
        Collects all the bets players made this round and adds them to
        the total amount. Resets player betting.
        """
        self.raised = 0
        for player_id, player_amount in self.player_amounts.items():
            self.amount += player_amount
            self.player_amounts[player_id] = 0

    def remove_player(self, player_id: int):
        """
        Removes the given player from the pot and adds their betting
        to the total amount.

        Arguments:
            player_id (int): Player player_id

        """
        if player_id not in self.player_amounts:
            return

        self.amount += self.player_amounts.pop(player_id)

    def get_amount(self) -> int:
        """
        Returns:
            int: How many chips this pot contains not including the current round of betting.

        """

        return self.amount

    def get_total_amount(self) -> int:
        """
        Returns:
            int: How many chips this pot contains including the current round of betting.

        """

        return sum(self.player_amounts.values()) + self.get_amount()

    def split_pot(self, raised_level: int) -> Optional[Pot]:
        """
        Splits the pot at the given raised level, and adds players with
        excess to the new pot.

        Arguments:
            raised_level (int)      - The chip count to cut off at
        Returns:
            (Optional[Pot])         - The new pot or None if self.raised <= raised_level
        """
        split_pot = Pot()

        if self.raised <= raised_level:
            return None

        # Overflow goes to split pot
        split_pot.raised = self.raised - raised_level
        self.raised = raised_level

        for player_id in self.players_in_pot():
            # player currently in last pot, post overflow to the split pot
            if self.get_player_amount(player_id) > self.raised:
                overflow = self.get_player_amount(player_id) - self.raised
                split_pot.player_post(player_id, overflow)
                self.player_amounts[player_id] -= -overflow

        return split_pot


class GameState(Enum):
    """An enum representing the state of the game (not just one hand)"""

    RUNNING = auto()
    """The table is active and is able to play hands."""

    STOPPED = auto()
    """The table is inactive due to lack of players or lack
    of chips and unable to play hands."""


class TexasHoldEm:
    # pylint: disable=too-many-instance-attributes,stop-iteration-return

    """
    Represents a table of TexasHoldEm (tournament style).

    Instantiate this object with the buyin, big blind, small blind,
    and the number of players.

    To interact with this class, call :meth:`TexasHoldEm.start_hand` which will
    run the PREHAND face (move/post blinds, reset pots, deal cards, etc.)

    To input an action at each stage, call :meth:`TexasHoldEm.take_action` which will
    execute the given action for the current player.

    """

    def __init__(self, buyin: int, big_blind: int, small_blind: int, max_players=9):
        """
        Represents a table of TexasHoldEm (tournament style).

        Instantiate this object with the buyin, big blind, small blind,
        and the number of players.

        To interact with this class, call :meth:`TexasHoldEm.run_hand` which returns an
        iterator that yields this object each stage of the game that requires input from
        a player.
        To input an action at each stage, call :meth:`TexasHoldEm.set_action` which will
        use the given action for the next state in the iterator.

        Arguments:
            buyin (int): The buyin to register for this game.
            big_blind (int): Big blind
            small_blind (int): Small blind
            max_players (int): how many players can sit at the table, defaults to 9.
        """
        self.buyin = buyin
        self.big_blind = big_blind
        self.small_blind = small_blind
        self.max_players = max_players

        self.players: tuple[Player] = tuple(Player(i, self.buyin) for i in range(max_players))

        self.btn_loc = random.randrange(0, max_players)
        self.bb_loc = -1
        self.sb_loc = -1
        self.current_player = -1

        self.pots = []
        self._deck = None
        self.board = []
        self.hands = {}

        self.num_hands = 0
        self.hand_phase = HandPhase.PREHAND
        self.game_state = GameState.RUNNING

        self._handstate_handler: Dict[HandPhase, Callable] = {
            HandPhase.PREHAND: self._prehand,
            HandPhase.PREFLOP: self._preflop,
            HandPhase.FLOP: self._flop,
            HandPhase.TURN: self._turn,
            HandPhase.RIVER: self._river,
            HandPhase.SETTLE: self._settle
        }

        self.hand_history: Dict[HandPhase, Union[PrehandHistory, BettingRoundHistory]] = {}
        self._action = None, None
        self._hand_gen = None

    def _prehand(self):
        """
        Handles skips, not enough chips, rotation and posting of blinds,
        dealing cards and setup for preflop.
        """
        if self.hand_phase != HandPhase.PREHAND:
            raise ValueError("Not time for prehand!")

        # set statuses for players
        for player_id in self.player_iter(loc=0):
            self.players[player_id].last_pot = 0

            if self.players[player_id].chips == 0:
                self.players[player_id].state = PlayerState.SKIP
            else:
                self.players[player_id].state = PlayerState.TO_CALL

        active_players = list(self.active_iter(self.btn_loc + 1))

        # stop if only 1 player
        if len(active_players) <= 1:
            self.game_state = GameState.STOPPED
            return

        # change btn loc (at least 2 players)
        self.btn_loc = active_players[0]
        self.sb_loc = active_players[1]

        # heads up edge case => sb = btn
        if len(active_players) == 2:
            self.sb_loc = self.btn_loc

        self.bb_loc = next(self.active_iter(self.sb_loc + 1))

        # reset pots
        self.pots = [Pot()]

        # reset hand history
        self.hand_history = {
            HandPhase.PREHAND: PrehandHistory(
                random_seed=random.getstate(),
                btn_loc=self.btn_loc,
                player_chips={
                    i: self.players[i].chips
                    for i in range(self.max_players)
                }
            )
        }
        self._action = None, None

        # post blinds
        self._player_post(self.sb_loc, self.small_blind)
        self._player_post(self.bb_loc, self.big_blind)

        # deal cards
        self._deck = Deck()
        self.hands = {}
        self.board = []

        for player_id in self.active_iter(self.btn_loc + 1):
            self.hands[player_id] = self._deck.draw(num=2)

        # action to left of BB
        self.current_player = next(self.active_iter(loc=self.bb_loc + 1))
        self.num_hands += 1

    def player_iter(self, loc: int = None) -> Iterator[int]:
        """
        Iterates through all players starting at player_id and rotating in order
        of increasing player_id.

        Arguments:
            loc (int)       - The player_id to start at, default is current_player.
        Returns:
            (Iterator[int]) - An iterator over all the players.
        """
        if loc is None:
            loc = self.current_player

        for i in range(loc, loc + self.max_players):
            yield i % self.max_players

    def active_iter(self, loc: int = None) -> Iterator[int]:
        """
        Iterates through all "active" players (i.e. all players without statuses
        OUT or SKIP).

        Arguments:
            loc (int)  - The location to start at, defaults to current_player
        Returns:
            (Iterator[int]) - An iterator over all active players starting at loc
        """
        if loc is None:
            loc = self.current_player
        for player_id in self.player_iter(loc=loc):
            if self.players[player_id].state not in (PlayerState.OUT, PlayerState.SKIP):
                yield player_id

    def in_pot_iter(self, loc: int = None) -> Iterator[int]:
        """
        Iterates through all active players, that can take an action.
        Iterates thru self._active_iter() and finds players with state
        IN or TO_CALL (i.e. not including ALL_IN).

        Arguments:
            loc (int)       - The location to start at, defaults to current_player
        Returns:
            (Iterator[int]) - An iterator over active players who can take an action.
        """
        if loc is None:
            loc = self.current_player
        for player_id in self.active_iter(loc=loc):
            if self.players[player_id].state in (PlayerState.IN, PlayerState.TO_CALL):
                yield player_id

    def _split_pot(self, pot_id: int, raised_level: int):
        """
        Splits the given pot at the given raised level, and adds players with
        excess to the new pot.

        Arguments:
            pot_id (int)            - The pot to split
            raised_level (int)      - The chip count to cut off at

        """
        pot = self._get_pot(pot_id)
        split_pot = pot.split_pot(raised_level)

        if not split_pot:
            return

        self.pots.insert(pot_id + 1, split_pot)

        # increment last_pot for players with enough chips
        for player_id in self.in_pot_iter():
            if self.players[player_id].chips >= self.chips_to_call(player_id):
                self.players[player_id].last_pot += 1

    def _player_post(self, player_id: int, amount: int):
        """
        Let a player post the given amount and sets the corresponding board state
        (i.e. makes other player states TO_CALL, sets ALL_IN). Also handles all
        pots (i.e. split pots).

        Arguments:
            player_id (int) - The player_id of the player posting
            amount (int)	- The amount to post
        """
        amount = min(self.players[player_id].chips, amount)
        original_amount = amount
        last_pot = self.players[player_id].last_pot
        chips_to_call = self._get_pot(last_pot).chips_to_call(player_id)

        # if a player posts, they are in the pot
        if amount == self.players[player_id].chips:
            self.players[player_id].state = PlayerState.ALL_IN
        else:
            self.players[player_id].state = PlayerState.IN

        # call in previous pots
        for i in range(last_pot):
            amount = amount - self._get_pot(i).chips_to_call(player_id)
            self.pots[i].player_post(player_id, self.pots[i].chips_to_call(player_id))

        self._get_pot(last_pot).player_post(player_id, amount)

        # players previously in pot need to call in event of a raise
        if amount > chips_to_call:
            for pot_player_id in self._get_pot(last_pot).players_in_pot():
                if self._get_pot(last_pot).chips_to_call(pot_player_id) > 0 and \
                   self.players[pot_player_id].state == PlayerState.IN:
                    self.players[pot_player_id].state = PlayerState.TO_CALL

        # if a player is all_in in this pot, split a new one off
        if PlayerState.ALL_IN in (self.players[i].state
                                  for i in self._get_pot(last_pot).players_in_pot()):
            raised_level = min(self._get_pot(last_pot).get_player_amount(i)
                               for i in self._get_pot(last_pot).players_in_pot()
                               if self.players[i].state == PlayerState.ALL_IN)
            self._split_pot(last_pot, raised_level)

        self.players[player_id].chips = self.players[player_id].chips - original_amount

    def _get_pot(self, pot_id: int) -> Pot:
        """
        Arguments:
            pot_id (int): The player_id of the pot to get
        Returns:
            Pot: The pot with given player_id
        Raises:
            ValueError: If a pot with player_id pot_id does not exist.

        """
        if pot_id >= len(self.pots):
            raise ValueError(f"Pot with player_id {pot_id} does not exist.")

        return self.pots[pot_id]

    def _get_last_pot(self) -> Pot:
        """
        Returns:
            Pot: The current "active" pot

        """
        return self._get_pot(self._last_pot_id())

    def _last_pot_id(self) -> int:
        """
        Returns:
            int: The pot player_id of the last pot.

        """
        return len(self.pots) - 1

    def _is_hand_over(self) -> bool:
        """
        Returns:
            bool: True if no more actions can be taken by the remaining players.
        """
        count = 0
        for i in self.in_pot_iter():
            if self.players[i].state == PlayerState.TO_CALL:
                return False

            if self.players[i].state == PlayerState.IN:
                count += 1

            if count > 1:
                return False
        return True

    def _preflop(self) -> Iterator[TexasHoldEm]:
        """
        Runs the PREFLOP sequence
        """
        if self.hand_phase != HandPhase.PREFLOP:
            raise ValueError("Not time for PREFLOP!")

        self.current_player = next(self.active_iter(loc=self.bb_loc + 1))
        yield from self._betting_round()

    def _flop(self) -> Iterator[TexasHoldEm]:
        """
        Runs the FLOP sequence
        """
        if self.hand_phase != HandPhase.FLOP:
            raise ValueError("Not time for PREFLOP!")

        self.board.extend(self._deck.draw(num=3))
        self.current_player = next(self.active_iter(loc=self.btn_loc + 1))
        yield from self._betting_round()

    def _turn(self) -> Iterator[TexasHoldEm]:
        """
        Runs the TURN sequence
        """
        if self.hand_phase != HandPhase.TURN:
            raise ValueError("Not time for Turn!")

        self.board.extend(self._deck.draw(num=1))
        self.current_player = next(self.active_iter(loc=self.btn_loc + 1))
        yield from self._betting_round()

    def _river(self) -> Iterator[TexasHoldEm]:
        """
        Runs the RIVER sequence
        """
        if self.hand_phase != HandPhase.RIVER:
            raise ValueError("Not time for River!")

        self.board.extend(self._deck.draw(num=1))
        self.current_player = next(self.active_iter(loc=self.btn_loc + 1))
        yield from self._betting_round()

    def _settle(self):
        """
        Settles the current hand. If players are all-in, makes sure
        the board has 5 card before settling.
        """
        if self.hand_phase != HandPhase.SETTLE:
            raise ValueError("Not time for Settle!")

        self.current_player = next(self.active_iter(loc=self.btn_loc + 1))

        for pot in self.pots:
            players_in_pot = list(pot.players_in_pot())
            # only player left in pot wins
            if len(players_in_pot) == 1:
                self.players[players_in_pot[0]].chips += pot.get_total_amount()
                continue

            # make sure there is 5 cards on the board
            if len(self.board) < 5:
                self.board.extend(self._deck.draw(num=5 - len(self.board)))

            player_ranks = {}
            for player_id in players_in_pot:
                player_ranks[player_id] = evaluator.evaluate(self.hands[player_id], self.board)

            best_rank = min(player_ranks.values())
            winners = [player_id
                       for player_id, player_rank in player_ranks.items()
                       if player_rank == best_rank]

            win_amount = (pot.get_total_amount()) / len(winners)
            win_amount = round(win_amount)
            for player_id in winners:
                self.players[player_id].chips += win_amount

    def chips_to_call(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int) - The player player_id
        Returns:
            int - The amount of chips the player needs to call in all pots
                  to play the hand.
        """
        return sum(self._get_pot(i).chips_to_call(player_id) for i in range(len(self.pots)))

    def player_bet_amount(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int) - The player player_id
        Returns:
            int - The amount of chips the player bet this round across all
                  pots.
        """
        return sum(self._get_pot(i).get_player_amount(player_id) for i in range(len(self.pots)))

    def chips_at_stake(self, player_id: int) -> int:
        """
        Arguments:
            player_id (int) - The player player_id
        Returns:
            int - The amount of chips the player is eligible to win
        """
        return sum(self._get_pot(i).get_total_amount()
                   for i in range(len(self.pots))
                   if player_id in self._get_pot(i).players_in_pot())

    def validate_move(self, player_id: int,
                      action: ActionType,
                      value: Optional[int] = None) -> bool:
        """
        Validate the potentially invalplayer_id action for the given player.

        Arguments:
            player_id (int): the player to take action
            action (ActionType): The ActionType to take
            value (int, optional): In the case of raise, how much to raise
        Returns:
            bool: True if the move is valplayer_id, False o/w
        """
        # ALL_IN should be translated
        new_action, new_value = action, value
        if new_action == ActionType.ALL_IN:
            new_action, new_value = self._translate_allin(new_action, new_value)

        player_amount = self.player_bet_amount(player_id)
        chips_to_call = self.chips_to_call(player_id)

        # Check if player player_id is current player
        if self.current_player != player_id:
            return False

        if new_action == ActionType.CALL:
            return self.players[player_id].state == PlayerState.TO_CALL
        if new_action == ActionType.CHECK:
            return self.players[player_id].state == PlayerState.IN
        if new_action == ActionType.RAISE:
            return not (
               new_value is None or
               (new_value < self.big_blind and action != ActionType.ALL_IN) or
               player_amount + self.players[player_id].chips < new_value or
               new_value < chips_to_call
            )
        if new_action == ActionType.FOLD:
            return True

        return False

    def _safe_execute(self, player_id: int,
                      action: ActionType,
                      value: Optional[int] = None) -> bool:
        """
        Safely execute the potentially invalplayer_id action for the given player.

        Arguments:
            player_id (int) 				- the player to take action
            action (ActionType) 	- The ActionType to take
            value (Optional[int]) - In the case of raise, how much to raise
        Returns:
            bool - True if successfully executed, False otherwise
        """
        # Validate move
        if not self.validate_move(player_id, action, value):
            return False

        # ALL_IN should be translated
        if action == ActionType.ALL_IN:
            action, value = self._translate_allin(action, value)

        player_amount = self.player_bet_amount(player_id)
        chips_to_call = self.chips_to_call(player_id)

        # Execute move
        if action == ActionType.CALL:
            self._player_post(player_id, chips_to_call)
        elif action == ActionType.CHECK:
            pass
        elif action == ActionType.RAISE:
            self._player_post(player_id, value - player_amount)
        elif action == ActionType.FOLD:
            self.players[player_id].state = PlayerState.OUT
            for i in range(self.players[player_id].last_pot + 1):
                self.pots[i].remove_player(player_id)
        else:
            return False

        return True

    def _translate_allin(self,
                         action: ActionType,
                         value: int = None) -> Tuple[ActionType, Optional[int]]:
        """
        Translates an all-in action to the appropriate action,
        either call or raise.
        """
        if action != ActionType.ALL_IN:
            return action, value

        if self.players[self.current_player].chips <= self.chips_to_call(self.current_player):
            return ActionType.CALL, None

        return ActionType.RAISE, \
            self.player_bet_amount(self.current_player) + self.players[self.current_player].chips

    def _betting_round(self) -> Iterator[TexasHoldEm]:
        """
        Core round of the poker game. Executes actions from each active player
        until everyone "checks"

        Raises:
            ValueError - If self.hand_state is not a valplayer_id betting round
        """

        if self.hand_phase in (HandPhase.PREHAND, HandPhase.SETTLE):
            raise ValueError("Not valplayer_id betting round!")

        first_pot = self._last_pot_id()
        player_iter = self.in_pot_iter(self.current_player)

        while not self._is_hand_over():
            try:
                self.current_player = next(player_iter)
            except StopIteration:
                break

            yield self

            action, val = self._translate_allin(*self._action)
            passed = self._safe_execute(self.current_player, action, val)

            if not passed:
                raise ValueError(f"Invalplayer_id move for player {self.current_player}: "
                                 f"{action}, {val}")

            if self.hand_phase not in self.hand_history:
                self.hand_history[self.hand_phase] = BettingRoundHistory(actions=[])
            betting_history = self.hand_history.get(self.hand_phase)
            betting_history.actions.append(PlayerAction(
                player_id=self.current_player,
                action_type=action,
                value=val)
            )

            # On raise, everyone eligible gets to take another action
            if action == ActionType.RAISE:
                player_iter = self.in_pot_iter(self.current_player)

                # Throwaway current player
                # Edge case: _in_pot_iter already excludes ALL_IN
                if self.players[self.current_player].state != PlayerState.ALL_IN:
                    next(player_iter)

        # consolidate betting to all pots in this betting round
        for i in range(first_pot, len(self.pots)):
            self._get_pot(i).collect_bets()

    def get_hand(self, player_id) -> list[Card]:
        """
        Arguments:
            player_id (int): The player_id of the player
        Returns:
            list[Card]: A two element list of the hand of the given player,
                if player has not been dealt a hand, returns an empty list

        """
        return self.hands.get(player_id, [])

    def start_hand(self):
        """
        Starts a new hand.

        Raises:
            (ValueError)            - If hand already in progress.
        """
        if self.is_hand_running():
            raise ValueError('In the middle of a hand!')

        self.hand_phase = HandPhase.PREHAND
        self._handstate_handler[self.hand_phase]()

        if self.game_state == GameState.STOPPED:
            return

        self.hand_phase = self.hand_phase.next_phase()
        self._hand_gen = self._hand_iter()

        try:
            next(self._hand_gen)
        except StopIteration:
            pass

    def take_action(self, action_type: ActionType, value: Optional[int] = None):
        """
        The current player takes the specified action.

        Arguments:
            action_type (ActionType) - The action type
            value (Optional[int])    - The value
        Raises:
            (ValueError)            - If no action can be taken due to GameState.STOPPED
                                      or if the move is invalid.

        """
        if not self.is_hand_running():
            raise ValueError("No hand is running")

        if not self.validate_move(self.current_player, action_type, value=value):
            raise ValueError("Move is invalid!")

        self._action = (action_type, value)

        try:
            next(self._hand_gen)
        except StopIteration:
            pass

    def _hand_iter(self) -> Iterator[TexasHoldEm]:
        """
        Returns:
            (Iterator[TexasHoldEm])	- A generator over every intermediate game state.
                                      i.e. right before every action.
        Raises:
            (ValueError)            - If phase != PREFLOP
        """
        if self.hand_phase != HandPhase.PREFLOP:
            raise ValueError("Cannot iterate over hand: not time for PREFLOP")

        while self.is_hand_running():
            if self._is_hand_over():
                self.hand_phase = HandPhase.SETTLE

            round_iter: Optional[Iterator[TexasHoldEm]] = self._handstate_handler[self.hand_phase]()
            if round_iter:
                yield from round_iter

            self.hand_phase = self.hand_phase.next_phase()

    def is_hand_running(self) -> bool:
        """
        Returns:
            bool: True if there is a hand running, false o/w

        """
        return self.hand_phase != HandPhase.PREHAND

    def is_game_running(self) -> bool:
        """
        Returns:
            bool: True if the game is running, false o/w

        """
        return self.game_state == GameState.RUNNING



from texasholdem.game.player_state import PlayerState
from texasholdem.game.game import HandPhase
from texasholdem.card import card
from texasholdem.game.action_type import ActionType


class TextGUI:
    def __init__(self):
        self.top_spaces = 125
        self.spaces_between = 2
        self.player_ids = []

    def set_player_ids(self, ids):
        self.player_ids = ids

    def _player_state_to_string(self, state):
        if state == PlayerState.SKIP:
            return "SKIP"
        elif state == PlayerState.OUT:
            return "OUT"
        elif state == PlayerState.IN:
            return "IN"
        elif state == PlayerState.TO_CALL:
            return "TO_CALL"
        elif state == PlayerState.ALL_IN:
            return "ALL_IN"

        return "UNKNOWN STATE"

    def _action_to_string(self, action, value):
        if action == ActionType.FOLD:
            return "folds"
        elif action == ActionType.RAISE:
            return "raises to {}".format(value)
        elif action == ActionType.CALL:
            return "calls"
        elif action == ActionType.CHECK:
            return "checks"
        elif action == ActionType.ALL_IN:
            return "goes all in"

    def _get_player_summary(self, poker_game, id):
        card_ids = self.player_ids
        if poker_game.hand_phase == HandPhase.SETTLE:
            card_ids = list(range(poker_game.max_players))

        lines = []

        lines.append("Player {}".format(id))

        if poker_game.btn_loc == id:
            lines.append("Button")
        elif poker_game.sb_loc == id:
            lines.append("Small Blind")
        elif poker_game.bb_loc == id:
            lines.append("Big Blind")

        lines.append("Chips: {}".format(poker_game.players[id].chips))

        status = self._player_state_to_string(poker_game.players[id].state)
        lines.append(status)
        if poker_game.players[id].state == PlayerState.SKIP:
            return lines

        if poker_game.is_hand_running() and id in card_ids:
            cards_int = poker_game.get_hand(id)
            if not cards_int:
                return lines

            cards = card.card_list_to_pretty_str(cards_int)
            lines.append("Cards: {}".format(cards))

        if poker_game.players[id].state == PlayerState.OUT:
            return lines

        bets = sum(poker_game._get_pot(i).get_player_amount(id) for i in range(poker_game.players[id].last_pot + 1))
        lines.append("Bet: {}".format(bets))

        return lines

    def _get_board_summary(self, poker_game):
        lines = []

        pretty_cards = card.card_list_to_pretty_str(poker_game.board)
        lines.append(pretty_cards)

        for i in range(len(poker_game.pots)):
            lines.append("Pot {}: {} ({})".format(i,
                                                  poker_game.pots[i].get_total_amount(),
                                                  poker_game.pots[i].get_amount()))

        return lines

    def accept_input(self):
        args = input("Your Turn!: ")

        if " " in args:
            action_str, val = args.split()
        else:
            action_str, val = args, 0
        action_str = action_str.lower()

        if action_str == "call":
            return ActionType.CALL, None
        elif action_str == "fold":
            return ActionType.FOLD, None
        elif action_str == "all-in":
            return ActionType.ALL_IN, None
        elif action_str == "raise":
            return ActionType.RAISE, float(val)
        elif action_str == "check":
            return ActionType.CHECK, None
        else:
            # always invalid
            return ActionType.RAISE, -1

    def print_action(self, id, action, val):
        text = "Player {} {}.".format(id, self._action_to_string(action, val))
        print(text)

    def print_state(self, poker_game):
        text = ""
        ordering = [[3], [2, 4], ["board"], [1, 5], [0]]

        for lst in ordering:
            n = len(lst)
            if n > 1:
                lines_lst = []
                for elem in lst:
                    lines_lst.append(self._get_player_summary(poker_game, elem))

                max_lines = max([len(x) for x in lines_lst])
                for lines in lines_lst:
                    length = len(lines)
                    for _ in range(0, max_lines - length):
                        lines.append("")

                lines = []
                if len(lines_lst) > 1:
                    for line_tuple in zip(*lines_lst):
                        agg_len = sum(len(line) for line in line_tuple)
                        space_num = self.top_spaces - agg_len
                        space_between = space_num // len(line_tuple)

                        result = ""
                        for j in range(len(line_tuple)):
                            result += line_tuple[j]
                            if j != len(line_tuple) - 1:
                                result += " " * space_between
                        lines.append(result)

            elif lst[0] == "board":
                lines = self._get_board_summary(poker_game)
            else:
                lines = self._get_player_summary(poker_game, lst[0])

            for i in range(len(lines)):
                space_num = (self.top_spaces - len(lines[i])) // 2
                lines[i] = " " * space_num + lines[i]

            for i in range(len(lines)):
                text += lines[i] + "\n"
            text += "\n" * (self.spaces_between - 1)

        print(text)



"""The handphase module contains an enum with the possible handphases
of the game. It includes the well-known PREHAND, PREFLOP, FLOP, and RIVER
phases, in addition to new PREHAND and SETTLE phases purely for book-keeping."""

from __future__ import annotations

from enum import Enum, auto


class HandPhase(Enum):
    """An enum representing the phase of the hand."""

    PREHAND = auto()
    """In this phase, players sit out if requested, players
    rejoin if requested, blinds are moved and posted, and card
    are dealt."""

    PREFLOP = auto()
    FLOP = auto()
    TURN = auto()
    RIVER = auto()

    SETTLE = auto()
    """In this phase, the pots are settled and the last aggressor
    shows their card, while every player in turn either shows
    or folds their card."""

    def next_phase(self) -> HandPhase:
        """
        Returns:
            HandPhase: The next HandPhase after this one
        """
        return _next_phase_dict[self]


_next_phase_dict = {
    HandPhase.PREHAND: HandPhase.PREFLOP,
    HandPhase.PREFLOP: HandPhase.FLOP,
    HandPhase.FLOP: HandPhase.TURN,
    HandPhase.TURN: HandPhase.RIVER,
    HandPhase.RIVER: HandPhase.SETTLE,
    HandPhase.SETTLE: HandPhase.PREHAND
}



"""The history module includes various dataclasses intended
to keep the history of the game to be able to replay hands and
save what happened. It includes a random_seed so that seeding
the python random module with it will run the hand exactly
as it occurred.
"""

from typing import Optional
from dataclasses import dataclass

from texasholdem.game.action_type import ActionType


@dataclass(frozen=True)
class PrehandHistory:
    """Prehand history class, includes random_seed, button location, and
    the player chip counts."""
    random_seed: object
    btn_loc: int
    player_chips: dict[int, int]


@dataclass(frozen=True)
class PlayerAction:
    """PlayerAction history class, includes the player id, the action type,
    and the value."""
    player_id: int
    action_type: ActionType
    value: Optional[int]


@dataclass(frozen=True)
class BettingRoundHistory:
    """BettingRound history class, includes a list of PlayerActions."""
    actions: list[PlayerAction]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['texasholdem',
 'texasholdem.card',
 'texasholdem.evaluator',
 'texasholdem.game',
 'texasholdem.gui']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'texasholdem',
    'version': '0.2.0',
    'description': 'A texasholdem python package',
    'long_description': '# texasholdem\nA python package for Texas Hold \'Em Poker.\n\n[Current Release Version v0.2.0](https://github.com/SirRender00/texasholdem/releases/tag/v0.2.0)\n\n[v1.0.0 Roadmap](https://github.com/SirRender00/texasholdem/wiki/Version-1.0.0-Roadmap)\n\n## Contributing\nTo be added as a contributor, please email me at evyn.machi@gmail.com with your GitHub username and mention one of the open issues / a new issue you would like to tackle first.\nFor more information about contributing, please see the wiki.\n\n## Install\n```bash\npip install texasholdem\n```\n\n## Quickstart Guide\nStarting a game is as simple as the following:\n```python\nfrom texasholdem import TexasHoldEm\n\ngame = TexasHoldEm(buyin=500, \n                   big_blind=5, \n                   small_blind=2,\n                   max_players=9)\ngame.start_hand()\nwhile game.is_hand_running():\n    game.take_action(...)\n```\n\n## Game Information\nGet game information and take actions through intuitive attributes:\n```python\nfrom texasholdem import TexasHoldEm, HandPhase, ActionType\n\ngame = TexasHoldEm(buyin=500, \n                   big_blind=5, \n                   small_blind=2,\n                   max_players=9)\ngame.start_hand()\n\nassert game.hand_phase == HandPhase.PREFLOP\nassert HandPhase.PREFLOP.next_phase() == HandPhase.FLOP\nassert game.chips_to_call(game.current_player) == game.big_blind\n\ngame.take_action(ActionType.CALL)\ngame.take_action(ActionType.RAISE, value=10)\n\nassert game.chips_to_call(game.current_player) == 10 - game.big_blind\n```\n\n## Card Module\nThe card module represents cards as 32-bit integers for simple and fast hand\nevaluations. For more information about the representation, see the `Card`\nmodule.\n\n```python\nfrom texasholdem.card import Card\n\ncard = Card("Kd")                       # King of Diamonds\nassert isinstance(card, int)            # True\nassert card.rank == 11                  # 2nd highest rank (0-12)\nassert card.pretty_string == "[ K ♦ ]"\n```\n\nThe `game.get_hand(player_id=...)` method of the `TexasHoldEm` class \nwill return a list of type `list[Card]`.\n\n## Evaluator Module\nThe evaluator module returns the rank of the best 5-card hand from a list of 5 to 7 cards.\nThe rank is a number from 1 (strongest) to 7462 (weakest). This determines the winner in the `TexasHoldEm` module:\n\n```python\nfrom texasholdem.card import Card\nfrom texasholdem.evaluator import evaluate, rank_to_string\n\nassert evaluate(cards=[Card("Kd"), Card("5d")],\n                board=[Card("Qd"), \n                       Card("6d"), \n                       Card("5s"), \n                       Card("2d"),\n                       Card("5h")]) == 927\nassert rank_to_string(927) == "Flush, King High"\n```\n',
    'author': 'Evyn Machi',
    'author_email': 'evyn.machi@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SirRender00/texasholdem',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""The player state module includes an enum which represents the player
state. For example, if a player is in the pot with the proper amount
of chips, that player is said to be IN. If a player needs to call a bet,
that player has status TO_CALL. If a player has no more chips to bet,
that player is ALL_IN, etc."""

from enum import Enum, auto


class PlayerState(Enum):
    """An enum representing a player state (i.e. needs to
    call, in the pot, sitting out, etc."""

    SKIP = auto()
    """Player is sitting out this hand, they will not be dealt
    cards and will rejoin upon request."""

    OUT = auto()
    """Player has folded their hand this round."""

    IN = auto()
    """Player is in the latest pot and has put in enough chips."""

    TO_CALL = auto()
    """Player is in the latest pot and needs to call a raise."""

    ALL_IN = auto()
    """Player is all-in and cannot take more actions this hand."""



"""
Evaluates hand strengths using a variant of Cactus Kev's algorithm:
http://www.suffecool.net/poker/evaluator.html

I make considerable optimizations in terms of speed and memory usage,
in fact the lookup table generation can be done in under a second and
consequent evaluations are very fast. Won't beat C, but very fast as
all calculations are done with bit arithmetic and table lookups.
"""

import itertools
from texasholdem.card import card
from texasholdem.card.card import Card
from texasholdem.evaluator.lookup_table import LOOKUP_TABLE


def _five(cards: list[Card]) -> int:
    """
    Performs an evaluation given card in integer form, mapping them to
    a rank in the range [1, 7462], with lower ranks being more powerful.

    Variant of Cactus Kev's 5 card evaluator.

    Args:
        cards (list[Card]): A list of 5 card ints.
    Returns:
        int: The rank of the hand.

    """
    # if flush
    if cards[0] & cards[1] & cards[2] & cards[3] & cards[4] & 0xF000:
        hand_or = (cards[0] | cards[1] | cards[2] | cards[3] | cards[4]) >> 16
        prime = card.prime_product_from_rankbits(hand_or)
        return LOOKUP_TABLE.flush_lookup[prime]

    # otherwise
    prime = card.prime_product_from_hand(cards)
    return LOOKUP_TABLE.unsuited_lookup[prime]


def _six(cards: list[Card]) -> int:
    """
    Calls `_five()` on all (6 choose 5) combinations of the input.
    Returns the max.

    Args:
        cards (list[Card]): A list of 6 card ints.
    Returns:
        int: The rank of the hand.

    """
    minimum = LOOKUP_TABLE.MAX_HIGH_CARD

    all5cardcombobs = itertools.combinations(cards, 5)
    for combo in all5cardcombobs:

        score = _five(combo)
        if score < minimum:
            minimum = score

    return minimum


def _seven(cards: list[Card]) -> int:
    """
    Calls `_five()` on all (7 choose 5) combinations of the input.
    Returns the max.

    Args:
        cards (list[Card]): A list of 7 card ints.
    Returns:
        int: The rank of the hand.

    """
    minimum = LOOKUP_TABLE.MAX_HIGH_CARD

    all5cardcombobs = itertools.combinations(cards, 5)
    for combo in all5cardcombobs:

        score = _five(combo)
        if score < minimum:
            minimum = score

    return minimum


_hand_size_map = {
    5: _five,
    6: _six,
    7: _seven
}


def evaluate(cards: list[Card], board: list[Card]) -> int:
    """
    Evaluates hand strengths using a variant of Cactus Kev's algorithm:
    http://www.suffecool.net/poker/evaluator.html

    Args:
        cards (list[int]): A list of length two of card ints that a player holds.
        board (list[int]): A list of length 3, 4, or 5 of card ints.
    Returns:
        int: A number between 1 (highest) and 7462 (lowest) representing the relative
            hand rank of the given card.

    """
    all_cards = cards + board
    return _hand_size_map[len(all_cards)](all_cards)


def _get_rank_class(hand_rank: int) -> int:
    """
    Returns the class of hand given the hand hand_rank
    returned from evaluate.

    """
    max_rank = min(rank
                   for rank in LOOKUP_TABLE.MAX_TO_RANK_CLASS
                   if hand_rank <= rank)
    return LOOKUP_TABLE.MAX_TO_RANK_CLASS[max_rank]


def rank_to_string(hand_rank: int) -> str:
    """
    Args:
        hand_rank (int): The rank of the hand given by :meth:`evaluate`
    Returns:
        string: A human-readable string of the hand rank (i.e. Flush, Ace High).

    """
    return LOOKUP_TABLE.RANK_CLASS_TO_STRING[_get_rank_class(hand_rank)]


def get_five_card_rank_percentage(hand_rank: int) -> float:
    """
    Args:
        hand_rank (int): The rank of the hand given by :meth:`evaluate`
    Returns:
        float: The percentile strength of the given hand_rank (i.e. what percent of hands is worse
            than the given one).

    """
    return 1 - float(hand_rank) / float(LOOKUP_TABLE.MAX_HIGH_CARD)
